"""Test flaskutils blueprint with Flask testing"""
import pathlib
from typing import Callable
from flask import Flask
import pytest

from conftest import BaseServer
import base_test

from h5grove.flaskutils import BLUEPRINT

# Fixtures ###


class _FlaskServer(BaseServer):
    """Class of objects provided by `flask_server` fixture"""

    def __init__(self, served_dir: pathlib.Path, client):
        super().__init__(served_dir)
        self.__client = client

    def _get_response(self, url: str, benchmark: Callable) -> BaseServer.Response:
        r = benchmark(lambda: self.__client.get(url))
        return BaseServer.Response(
            status=r.status_code, headers=list(r.headers), content=r.get_data()
        )


@pytest.fixture(scope="session")
def flask_server(tmp_path_factory):
    """Flask test client-based `server` fixture.

    Provides a function to fetch endpoints from the server.
    """
    base_dir = tmp_path_factory.mktemp("h5grove_flask_served").absolute()

    app = Flask("Test server")
    app.config["H5_BASE_DIR"] = str(base_dir)
    app.register_blueprint(BLUEPRINT)

    with app.test_client() as client:
        yield _FlaskServer(base_dir, client)


# Tests ###


class TestFlaskEndpoints(base_test.BaseTestEndpoints):
    """Test Flask blueprint endpoints using Flask test client"""

    @pytest.fixture
    def server(self, flask_server):
        yield flask_server



import pytest
import base_test


class TestExampleAppEndpoints(base_test.BaseTestEndpoints):
    """Test endpoints of example app run as subprocess"""

    @pytest.fixture
    def server(self, subprocess_server):
        yield subprocess_server



#!/usr/bin/env python
# coding: utf-8
"""Flask-based server sample code"""
import argparse
from flask import Flask
from flask_compress import Compress  # type: ignore
from flask_cors import CORS  # type: ignore
import os
from h5grove.flaskutils import BLUEPRINT as h5grove_blueprint


parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-p", "--port", type=int, default=8888, help="Port the server is listening on"
)
parser.add_argument("--ip", default="localhost", help="IP the server is listening on")
parser.add_argument(
    "--compression", action="store_true", help="Enable HTTP compression"
)
parser.add_argument(
    "--basedir", default=".", help="Base directory from which to retrieve HDF5 files"
)
options = parser.parse_args()

# Create Flask application
app = Flask(__name__)

# Enable cross-origin resource sharing, see https://flask-cors.readthedocs.io
CORS(app)

# Enable compression, see https://github.com/colour-science/flask-compress
if options.compression:
    Compress(app)

# Configure h5grove default endpoints
app.config["H5_BASE_DIR"] = os.path.abspath(options.basedir)
app.register_blueprint(h5grove_blueprint)

# Start server
app.run(port=options.port, host=options.ip)



from typing import Dict, Generic, Optional, Sequence, TypeVar, Union
import h5py
import numpy as np
import os

try:
    import hdf5plugin  # noqa: F401
except ImportError:
    pass
from .utils import attr_metadata, get_entity_from_file, parse_slice, sorted_dict


class EntityContent:
    type = "other"

    def __init__(self, path: str):
        self._path = path

    def metadata(self) -> Dict[str, str]:
        return {"name": self.name, "type": self.type}

    @property
    def name(self) -> str:
        return self._path.split("/")[-1]


class ExternalLinkContent(EntityContent):
    type = "externalLink"

    def __init__(self, path: str, link: h5py.ExternalLink):
        super().__init__(path)
        self._target_file = link.filename
        self._target_path = link.path

    def metadata(self, depth=None):
        return sorted_dict(
            ("target_file", self._target_file),
            ("target_path", self._target_path),
            *super().metadata().items(),
        )


class SoftLinkContent(EntityContent):
    type = "softLink"

    def __init__(self, path: str, link: h5py.SoftLink) -> None:
        super().__init__(path)
        self._target_path = link.path

    def metadata(self, depth=None):
        return sorted_dict(
            ("target_path", self._target_path), *super().metadata().items()
        )


T = TypeVar("T", h5py.Dataset, h5py.Datatype, h5py.Group)


class ResolvedEntityContent(EntityContent, Generic[T]):
    def __init__(self, path: str, h5py_entity: T):
        super().__init__(path)
        self._h5py_entity = h5py_entity

    def attributes(self, attr_keys: Sequence[str] = None):
        if attr_keys is None:
            return dict((*self._h5py_entity.attrs.items(),))

        return dict((key, self._h5py_entity.attrs[key]) for key in attr_keys)

    def metadata(self, depth=None):
        attribute_names = sorted(self._h5py_entity.attrs.keys())
        return sorted_dict(
            (
                "attributes",
                [
                    attr_metadata(self._h5py_entity.attrs.get_id(k))
                    for k in attribute_names
                ],
            ),
            *super().metadata().items(),
        )


class DatasetContent(ResolvedEntityContent[h5py.Dataset]):
    type = "dataset"

    def metadata(self, depth=None):
        return sorted_dict(
            ("dtype", self._h5py_entity.dtype.str),
            ("shape", self._h5py_entity.shape),
            *super().metadata().items(),
        )

    def data(self, selection: str = None):
        if selection is None:
            return self._h5py_entity[()]

        parsed_slice = parse_slice(self._h5py_entity, selection)

        return self._h5py_entity[parsed_slice]

    def data_stats(
        self, selection: str = None
    ) -> Union[Dict[str, None], Dict[str, Union[float, int]]]:
        data = self._get_finite_data(selection)

        if data.size == 0:
            return {
                "min": None,
                "max": None,
                "mean": None,
                "std": None,
            }

        cast = float if np.issubdtype(data.dtype, np.floating) else int
        return {
            "min": cast(np.min(data)),
            "max": cast(np.max(data)),
            "mean": cast(np.mean(data)),
            "std": cast(np.std(data)),
        }

    def _get_finite_data(self, selection: Optional[str]) -> np.ndarray:
        data = np.array(self.data(selection), copy=False)  # So it works with scalars

        if not np.issubdtype(data.dtype, np.floating):
            return data

        mask = np.isfinite(data)
        if np.all(mask):
            return data

        return data[mask]


class GroupContent(ResolvedEntityContent[h5py.Group]):
    type = "group"

    def __init__(self, path: str, h5py_entity: h5py.Group, h5file: h5py.File):
        super().__init__(path, h5py_entity)
        self._h5file = h5file

    def _get_child_metadata_content(self, depth=0):
        return [
            create_content(self._h5file, os.path.join(self._path, child_path)).metadata(
                depth
            )
            for child_path in self._h5py_entity.keys()
        ]

    def metadata(self, depth=1):
        if depth == 0:
            return super().metadata()

        return sorted_dict(
            ("children", self._get_child_metadata_content(depth - 1)),
            *super().metadata().items(),
        )


def create_content(h5file: h5py.File, path: Optional[str]):
    if path is None:
        path = "/"

    entity = get_entity_from_file(h5file, path)

    if isinstance(entity, h5py.ExternalLink):
        return ExternalLinkContent(path, entity)

    if isinstance(entity, h5py.SoftLink):
        return SoftLinkContent(path, entity)

    if isinstance(entity, h5py.Dataset):
        return DatasetContent(path, entity)

    if isinstance(entity, h5py.Group):
        return GroupContent(path, entity, h5file)

    if isinstance(entity, h5py.Datatype):
        return ResolvedEntityContent(path, entity)

    raise TypeError(f"h5py type {type(entity)} not supported")



import h5py
from numbers import Number
import numpy as np
from typing import Any, Sequence, Tuple, Union
from .models import H5pyEntity


def attr_metadata(attrId: h5py.h5a.AttrID) -> dict:
    return {"dtype": attrId.dtype.str, "name": attrId.name, "shape": attrId.shape}


def get_entity_from_file(h5file: h5py.File, path: str) -> H5pyEntity:
    if path == "/":
        return h5file[path]

    link = h5file.get(path, getlink=True)
    if isinstance(link, h5py.ExternalLink) or isinstance(link, h5py.SoftLink):
        try:
            return h5file[path]
        except (OSError, KeyError):
            return link

    return h5file[path]


def parse_slice(dataset: h5py.Dataset, slice_str: str) -> Tuple[Union[slice, int], ...]:
    if "," not in slice_str:
        return (parse_slice_member(slice_str, dataset.shape[0]),)

    slice_members = slice_str.split(",")

    if len(slice_members) > dataset.ndim:
        raise TypeError(
            f"{slice_str} is a {len(slice_members)}d slice while the dataset is {dataset.ndim}d"
        )

    return tuple(
        parse_slice_member(s, dataset.shape[i]) for i, s in enumerate(slice_members)
    )


def parse_slice_member(slice_member: str, max_dim: int) -> Union[slice, int]:
    if ":" not in slice_member:
        return int(slice_member)

    slice_params = slice_member.split(":")
    if len(slice_params) == 2:
        start, stop = slice_params

        return slice(
            int(start) if start != "" else 0, int(stop) if stop != "" else max_dim
        )

    if len(slice_params) == 3:
        start, stop, step = slice_params

        return slice(
            int(start) if start != "" else 0,
            int(stop) if stop != "" else max_dim,
            int(step) if step != "" else 1,
        )

    raise TypeError(f"{slice_member} is not a valid slice")


def sorted_dict(*args: Tuple[str, Any]):
    return dict(sorted(args))


def _sanitize_dtype(dtype: np.dtype) -> np.dtype:
    """Convert dtype to a dtype supported by js-numpy-parser.

    See https://github.com/ludwigschubert/js-numpy-parser

    :raises ValueError: For unsupported array dtype
    """
    if dtype.kind not in ("f", "i", "u"):
        raise ValueError("Unsupported array type")

    # Convert to little endian
    result = dtype.newbyteorder("<")

    if result.kind == "i" and result.itemsize > 4:
        return np.dtype("<i4")  # int64 -> int32

    if result.kind == "u" and result.itemsize > 4:
        return np.dtype("<u4")  # uint64 -> uint32

    if result.kind == "f" and result.itemsize < 4:
        return np.dtype("<f4")

    if result.kind == "f" and result.itemsize > 8:
        return np.dtype("<f8")

    return result


def sanitize_array(array: Sequence[Number], copy: bool = True) -> np.ndarray:
    """Ensure array save as .npy can be read back by js-numpy-parser.

    See https://github.com/ludwigschubert/js-numpy-parser

    :param array: Array to sanitize
    :param copy: Set to False to avoid copy if possible
    :raises ValueError: For unsupported array dtype
    """
    ndarray = np.array(array, copy=False)
    return np.array(ndarray, copy=copy, order="C", dtype=_sanitize_dtype(ndarray.dtype))



"""Helpers for usage with `Flask <https://flask.palletsprojects.com/>`_"""
from flask import Blueprint, current_app, request, Response
import h5py
import os
from typing import Any, Callable, Mapping, Optional

from .content import create_content, DatasetContent, ResolvedEntityContent
from .encoders import encode


__all__ = ["attr_route", "data_route", "meta_route", "URL_RULES", "BLUEPRINT"]


def make_encoded_response(content, format: Optional[str]) -> Response:
    """Prepare flask Response according to format"""
    encoded_content, headers = encode(content, format)
    response = Response(encoded_content)
    response.headers.update(headers)
    return response


def attr_route(file_path: str):
    """`/attr/<file_path>` endpoints handler"""
    filename = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    path = request.args.get("path")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = create_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)
        return make_encoded_response(content.attributes(), format)


def data_route(file_path: str):
    """`/data/<file_path>` endpoints handler"""
    filename = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data(selection), format)


def meta_route(file_path: str):
    """`/meta/<file_path>` endpoints handler"""
    filename = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    path = request.args.get("path")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = create_content(h5file, path)
        return make_encoded_response(content.metadata(), format)


def stats_route(file_path: str):
    """`/stats/<file_path>` endpoints handler"""
    filename = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data_stats(selection), format)


URL_RULES = {
    "/attr/<path:file_path>": attr_route,
    "/data/<path:file_path>": data_route,
    "/meta/<path:file_path>": meta_route,
    "/stats/<path:file_path>": stats_route,
}
"""Mapping of Flask URL endpoints to handlers"""


BLUEPRINT = Blueprint("h5grove", __name__)
"""Blueprint of h5grove endpoints.

It relies on `H5_BASE_DIR` being defined in the app config.
"""


def _init_blueprint(
    blueprint: Blueprint, url_rules: Mapping[str, Callable[[str], Any]]
):
    for url, view_func in url_rules.items():
        blueprint.add_url_rule(url, view_func=view_func)


_init_blueprint(BLUEPRINT, URL_RULES)



"""Helpers for usage with `Tornado <https://www.tornadoweb.org>`_"""
import os
from typing import Optional
import h5py
import tornado.web
from .content import DatasetContent, ResolvedEntityContent, create_content
from .encoders import encode


__all__ = [
    "BaseHandler",
    "AttributeHandler",
    "DataHandler",
    "MetadataHandler",
    "get_handlers",
]


class BaseHandler(tornado.web.RequestHandler):
    def initialize(self, base_dir, allow_origin=None) -> None:
        self.base_dir = base_dir
        self.allow_origin = allow_origin

    def get(self, file_path):
        path = self.get_query_argument("path", None)
        format = self.get_query_argument("format", None)

        with h5py.File(os.path.join(self.base_dir, file_path), "r") as h5file:
            content = self.get_content(h5file, path)

        encoded_content_chunks, headers = encode(content, format)

        for key, value in headers.items():
            self.set_header(key, value)
        for chunk in encoded_content_chunks:
            self.write(chunk)
        self.finish()

    def get_content(self, h5file, path):
        raise NotImplementedError

    def prepare(self):
        if self.allow_origin is not None:
            self.set_header("Access-Control-Allow-Origin", self.allow_origin)


class AttributeHandler(BaseHandler):
    def get_content(self, h5file, path):
        content = create_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)
        return content.attributes()


class DataHandler(BaseHandler):
    def get_content(self, h5file, path):
        selection = self.get_query_argument("selection", None)

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data(selection)


class MetadataHandler(BaseHandler):
    def get_content(self, h5file, path):
        content = create_content(h5file, path)
        return content.metadata()


class StatisticsHandler(BaseHandler):
    def get_content(self, h5file, path):
        selection = self.get_query_argument("selection", None)

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data_stats(selection)


def get_handlers(base_dir: Optional[str], allow_origin: Optional[str] = None):
    """Returns list of `Rule` arguments"""
    init_args = {"base_dir": base_dir, "allow_origin": allow_origin}
    return [
        (r"/attr/(.*)", AttributeHandler, init_args),
        (r"/data/(.*)", DataHandler, init_args),
        (r"/meta/(.*)", MetadataHandler, init_args),
        (r"/stats/(.*)", StatisticsHandler, init_args),
    ]
